#!/usr/bin/env perl
#===============================================================================
#
#         FILE: 98_TadoAPI.pm
#
#         USAGE: Module for FHEM 
#         Info: Turn $debug on for debugging
# 
# REQUIREMENTS: Below modules should be pre-installed.
# 				HTTP::Request::Common
# 				HTTP::Headers
#			  	Data::Dumper;
# 				JSON
#
#         BUGS: ---
#        NOTES: ---
#       AUTHOR: Philipp Wolfmajer
# ORGANIZATION:
#      VERSION: 1.0
#      CREATED: 04/12/2019 07:55:44 PM
#     REVISION: 11/21/2019 05:17:22 PM
#===============================================================================
package main;
use strict;
use warnings;
use utf8;
use FHEM::Meta;

use HTTP::Request::Common qw (POST GET PUT);
use HTTP::Headers;
use JSON;

####DEFAULTS############
my $client_id='public-api-preview';
my $client_secret='4HJGRffVR8xb3XdEUQpjgZ1VplJi6Xgw';
my $scope='home.user';
my $AuthURL = qq{https://auth.tado.com/oauth/token};
my $DataURL = qq{https://my.tado.com/api/v2/me};
my $QueryURL = qq{https://my.tado.com/api/v2/homes};
my $tokenFile = "./FHEM/FhemUtils/TadoAPI_token";
my $debug = 0;
my $header = {};
my $data = {};
my $TokenData = {};

# helpers
my $apiStatus = 1;

my %sets = ( 
  "getTemperature" => "", 
  "refreshToken" => "noArg",
  "password"     => "",
  "update" => "noArg",
  "setGeo" => "",
  "setZoneOverlay" => "",
  "setAllOverlays" => ""
 );

my %gets = (  
  "getZoneDevices" => "noArg",
  "getZoneInfo" => "noArg",
  "getGeo" => "",
  "getXTest" => "",
  "getMobileDevices" => "noArg"
);


sub
TadoAPI_Initialize($)
{
  my ($hash) = @_;

  $hash->{DefFn}     = "TadoAPI_Define";
  $hash->{InitFn}    = "TadoAPI_Init";
  $hash->{SetFn}     = "TadoAPI_Set";
  $hash->{GetFn}     = "TadoAPI_Get";
  $hash->{AttrList}  = 
  "homeID " .
  "mobileID " .
  "debug:1,0 " .
  $main::readingFnAttributes;

}

sub TadoAPI_Init($$)
{
  my ($hash,$args) = @_;
  my $u = "wrong syntax: define <name> TadoAPI <username> <homeID> [<mobileID>]";
  return $u if(int(@$args) < 2);  
return undef;
}

sub TadoAPI_Define($$)
{
  my ($hash, $def) = @_;
  my @a    = split( "[ \t]+", $def );
  my $name = shift @a;
  my $type = shift @a;
  my $tokenFileName = $tokenFile."_".$name;

  return "Invalid number of arguments: "
      . "define <name> TadoAPI <username> [<homeID>]"
  if ( int(@a) < 1 );

  my ( $user, $homeID ) = @a;
  Log3 $name, 3, "TadoAPI_Define $name: called ";
  $hash->{STATE}="defined";

  # Initialize the device
  return $@ unless ( FHEM::Meta::SetInternals($hash) );

  $hash->{TADO_USER} = $user;
  $hash->{TOKEN_FILE} = $tokenFileName;

  my @args = ($homeID);

  if ($main::init_done) {
    # do something?    
    
    return TadoAPI_Catch($@) if $@;
  }

  my $password = TadoAPI_readPassword($name);

  if (defined($password)){
    TadoAPI_LoadToken($hash);

    # start the status update timer
    RemoveInternalTimer($hash);
    InternalTimer( gettimeofday() + 10, "TadoAPI_Update", $hash, 0 );

    if ( defined($homeID) && $homeID ne "" ) {
        $attr{$name}{homeID} = $homeID; 
    } else{
      my $id = TadoAPI_GetHomeId($hash);
      if ( defined($id) && $id ne "" ) {
        $attr{$name}{homeID} = $id; 
      }
    }
  }
  return undef;
}

sub TadoAPI_Set(@) {
  my ($hash, @a) = @_;
  return "Need at least one parameters" if(@a < 2);
  my $cmd = $a[1];
  my $value = $a[2];
  my $name = $hash->{NAME};
  my $subcmd;

  #debug
  $debug = $attr{$name}{debug};
  if(!defined($sets{$cmd})) {
  	my @cmds = ();
    foreach my $key (sort keys %sets) {
      push @cmds, $sets{$key} ? $key.":".join(",",$sets{$key}) : $key;
    }
    return "Unknown argument $a[1], choose one of " . join(" ", @cmds);
  }

  if( $cmd eq 'setGeo' ) {
    return "Need at least two parameters (mobileID, Setting)" if(@a < 4);
      if(  $a[3] eq "on" ) {
        Log3 $name, 3, "TadoAPI: set $name: processing ($cmd)";
        TadoAPI_SetGeoById($hash, $value, 1);
      } else {
        Log3 $name, 3, "TadoAPI: set $name: processing ($cmd)";
        TadoAPI_SetGeoById($hash, $value, 0);
      }
      TadoAPI_GetGeoById($hash, $value);
      return undef;

    } elsif( $cmd eq 'setZoneOverlay' ) {
      Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
      return "Need at least two parameters (ZoneID, Setting)" if(@a < 4);
      
      if( $a[3] eq "off" ) {
        TadoAPI_SetZoneOverlayById($hash, $value, "off");
      } elsif ($a[3] > 1) {
        TadoAPI_SetZoneOverlayById($hash, $value, $a[3]);
      }
      Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
      return undef; 

    } elsif( $cmd eq 'setAllOverlays' ) {
      Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
      return "Need at least one parameter (Setting)" if(@a < 3);
      if( $value eq "off" ) {
        TadoAPI_SetAllOverlays($hash, "off");
      } elsif ($value > 1) {
        TadoAPI_SetAllOverlays($hash, $value);
      }
      Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
      return undef;     

    } elsif( $cmd eq 'refreshToken' ) {
        Log3 $name, 3, "TadoAPI: set $name: processing ($cmd)";
        RemoveInternalTimer($hash);
        InternalTimer( gettimeofday() + 10, "TadoAPI_Update", $hash, 0 );
        TadoAPI_LoadToken($hash);
      return undef;

    } elsif( $cmd eq 'update' ) {
        Log3 $name, 3, "TadoAPI: set $name: processing ($cmd)";
        TadoAPI_UpdateFn($hash);
      return undef;

    } elsif( $cmd eq 'getTemperature' ) {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        return "ZoneID as parameter needed" if (!$value);
        if( $value >= 1 ) {
          TadoAPI_GetZoneReadingsById($hash, $value);       
      } else {
        return "Wrong ZoneID";
      }
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
      return undef;
      

    } elsif( $cmd eq 'password' ) {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        # name und cmd überspringen
        shift @a;
        shift @a;
        # den Rest der das passwort enthält, als ein String
        $subcmd = join(" ",@a);        
        return TadoAPI_storePassword($name,$subcmd);
        #Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        # start the status update timer
        RemoveInternalTimer($hash);
        InternalTimer( gettimeofday() + 10, "TadoAPI_Update", $hash, 0 );
      return undef;
      }

    return TadoAPI_Catch($@) if $@;
  }

sub TadoAPI_Get(@) {
  my ($hash, @a) = @_;
  return "Need at least one parameters" if(@a < 2);
  my $cmd = $a[1];
  my $value = $a[2];
  my $name = $hash->{NAME};  
  my $homeID = $attr{$name}{homeID};
  my $message = undef;

  #debug
  $debug = $attr{$name}{debug};
  if(!defined($gets{$cmd})) {
  	my @cmds = ();
    foreach my $key (sort keys %gets) {
      push @cmds, $gets{$key} ? $key.":".join(",",$gets{$key}) : $key;
    }
    return "Unknown argument $a[1], choose one of " . join(" ", @cmds);
  }

  if($cmd =~ /\Qget\E/) {
    
    COMMAND_HANDLER: {
      
      $cmd eq "getGeo" and do {
        return "Need at least one parameter (mobileID)" if(@a < 3);
        return "Wrong MobileID" if (length($value) < 6);
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        TadoAPI_GetGeoById($hash, $value);
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        last;
      };

      $cmd eq "getMobileDevices" and do {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        my @data = TadoAPI_GetMobileDevices($hash);
        $message = "Device List:\n";
        foreach my $item ( @data ){
          print "\n";
          $message .= $item->{'name'} . ": " . $item->{'id'} . "\n";
        };
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        last;
      };

      $cmd eq "getXTest" and do {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        $message = "Name: " . TadoAPI_GetZoneNameById($hash, $value);
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        last;
      };

      $cmd eq "getZoneDevices" and do {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        my @devArr = TadoAPI_RequestTadoDevices($hash);
        my $devicecount = 0;
        $message = "Tado-Device(s):\n";
        for (my $i=0; $i < @devArr; $i++){
          my $tadodevices = $devArr[$i]->{'devices'};
          $message .= "ZoneID:" . ($i+1);
          my $spacer = 0;
          foreach my $item ( @$tadodevices ){
            $message .= "\t" if ($spacer > 0);
            $message .= " " . $item->{'serialNo'} . " Battery: " . $item->{'batteryState'} . "\n";
            $devicecount++;
            $spacer++;
          }
        }
        $message .= "There are $devicecount Tado-Device(s) in this HomeID(" . $attr{$name}{homeID} . ")." ;
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        last;
      };

      $cmd eq "getZoneInfo" and do {
        Log3 $name, 3, "TadoAPI $name" . ": " . "processing ($cmd)" if $debug;
        TadoAPI_GetZoneInfo($hash);
        my $zonecount = TadoAPI_GetZoneCount($hash);
        $message = "You have $zonecount Zones in Home " . $attr{$name}{homeID} . ".\n";
        for (my $i=1; $i <= $zonecount; $i++) {
          $message .= "Zone ID:$i: " . TadoAPI_GetZoneNameById($hash, $i) . "\n"; 
        }
        $message .= "See Logfile for more Info";
        Log3 $name, 3, "TadoAPI $name" . ": " . "$cmd finished\n";
        last;
      };
    }
  
    return $message if $message;
    return TadoAPI_Catch($@) if $@;
    return undef;
  }
}

sub TadoAPI_Catch($) {
  my $exception = shift;
  if ($exception) {
    $exception =~ /^(.*)( at.*FHEM.*)$/;
    return $1;
  }
  return undef;
}

sub TadoAPI_Undefine($$) {
    my ( $hash, $name ) = @_;

    RemoveInternalTimer($hash);

    #todo remove tokenfile

    return undef;
}

sub TadoAPI_OnlineStatus(@){
  my ($hash) = @_;
  my $name = $hash->{NAME};

  # test api status
  my $param = {
      url        => $AuthURL,
			timeout    => 5,
			hash       => $hash,
			method     => "GET",
			header     => "",  
		  callback   => \&TadoAPI_callback 
		};
  #test if api is reachable
  HttpUtils_NonblockingGet($param);
  
  return undef;
}

sub TadoAPI_LoadToken(@){
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $tokenFileName = $tokenFile."_".$name;
  my $tokenLifeTime = $hash->{TOKEN_LIFETIME};
  $tokenLifeTime = 0 if(!defined $tokenLifeTime || $tokenLifeTime eq '');
  my $Token = undef;
  #debug
  $debug = $attr{$name}{debug};

  TadoAPI_OnlineStatus($hash);

  if($apiStatus){
    eval {
      open(TOKENFILE, '<', $tokenFileName) or die("ERROR: $!");
      $Token = decode_json(<TOKENFILE>)};
    if($@ || $tokenLifeTime < gettimeofday()){    
      Log3 $name, 3, "TadoAPI $name" . ": " . "Error while loading: $@" if $debug && $@;
      Log3 $name, 3, "TadoAPI $name" . ": " . "No Token or Token is expired, requesting new one" if $debug && $tokenLifeTime < gettimeofday();
      TadoAPI_NewTokenRequest($hash);    
    }else{
      Log3 $name, 3, "TadoAPI $name" . ": " . "Token expires at " . localtime($tokenLifeTime) if $debug;
      # if token is about to expire, refresh him
      if(($tokenLifeTime-15) < gettimeofday()){
        Log3 $name, 3, "TadoAPI $name" . ": " . "Token will expire soon, refreshing" if $debug;
        TadoAPI_TokenRefresh($hash);
      }
    }
    close(TOKENFILE);
    return $Token;  
    }
return undef;
}

sub TadoAPI_NewTokenRequest(@) {
    my ($hash) = @_;
    my $name = $hash->{NAME};
    my $username = $hash->{TADO_USER};
    my $password = TadoAPI_readPassword($name);
    my $tokenFileName = $tokenFile."_".$name;

    Log3 $name, 3, "TadoAPI $name" . ": " . "calling NewTokenRequest()" if $debug;

    $data = {
      client_id => $client_id,
      client_secret => $client_secret,
      username => $username,
      password => $password,
      scope => $scope,
      grant_type=>'password'
    };

    my $param = {
      url           => $AuthURL,
      method        => 'POST',
      timeout       =>  5,
      hash          => $hash,
      data          => $data
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);

    if($err ne "")
    {
        Log3 $name, 3, "TadoAPI $name" . ": " . "NewTokenRequest: Error while requesting ".$param->{url}." - $err";
    }
    elsif($data ne "")
    {
        Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
        my $decoded_data = eval { decode_json($data) };
        if ($@){
          Log3 $name, 3, "TadoAPI $name" . ": " . "NewTokenRequest: decode_json failed, invalid json. error: $@ ";
        }else{
          #write token data in file
          open(TOKENFILE,">$tokenFileName") or die("ERROR: $!");
          print TOKENFILE $data . "\n";
          close(TOKENFILE);

          # token lifetime management
          $hash->{TOKEN_LIFETIME}  = gettimeofday() + $decoded_data->{'expires_in'};
          $hash->{TOKEN_LIFETIME_HR}  = localtime($hash->{TOKEN_LIFETIME});
          Log3 $name, 3, "TadoAPI $name" . ": " . "Retrived new authentication token successfully. Valid until " . localtime($hash->{TOKEN_LIFETIME})if $debug;
          $hash->{STATE}="reachable";
        }
    }
}

sub TadoAPI_TokenRefresh(@) {
    my ($hash) = @_;
    my $name = $hash->{NAME};

    Log3 $name, 3, "TadoAPI $name" . ": " . "calling TokenRefresh()" if $debug;

    # load token
    eval {
      open(TOKENFILE, '<', $hash->{TOKEN_FILE}) or die("ERROR: $!");
      $TokenData = decode_json(<TOKENFILE>)};

    $data = {
      client_id => $client_id,
      client_secret => $client_secret,
      scope => $scope,
      grant_type=>'refresh_token',
      refresh_token => $TokenData->{'refresh_token'}
    };

    my $request = {
			url           => $AuthURL,
			method        => 'POST',
			timeout       =>  5,
			callback      => \&Tado_StoreTokenCallback,
			hash          => $hash,
			data          => $data
		};

    Log3 $name, 3, "TadoAPI $name" . ": " . "Calling refresh token...";
    HttpUtils_NonblockingGet($request);
}

sub TadoAPI_Update(@){
  my ($hash) = @_;
  my $name = $hash->{NAME};

  Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_Update called" if $debug;

  # timer loop
  #
  
  my $nextTimer = "none";
  # if api online, try again in 5 minutes
  if ( $apiStatus ) {
      $nextTimer = gettimeofday() + 300;
  }

  RemoveInternalTimer($hash);
  InternalTimer( $nextTimer, "TadoAPI_Update", $hash, 0 );

  # update subs
	TadoAPI_UpdateFn($hash);

  return undef;
}

######################## tado methods ########################
##############################################################

sub TadoAPI_SetZoneOverlayById(@){
  my ($hash, $zoneID, $setting) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID}; 
	my $URL = $QueryURL . qq{/$homeID/zones/$zoneID/overlay};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $method = "";
    my $myjson ="";
	 
    Log3 $name, 3, "TadoAPI $name" . ": SetOverlay for Zone $zoneID (Setting: " . $setting . ") - " . "query-URL: $URL" if $debug; 

    if ($setting eq "off"){
        $method = "DELETE";
        Log3 $name, 3, "TadoAPI $name" . ": " . "Deleting Overlay for Zone $zoneID";
    }elsif($setting > 10){
      $method = "PUT";
      $myjson = { 
        setting => { 
            type => "HEATING",
            power => "ON",
            temperature => {
                celsius => $setting 
            },
          },
        termination => {
            type => "MANUAL"
          },
        };
        $myjson = encode_json($myjson);
    }

    my $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => $method,
			timeout       =>  3,
			callback      => \&Tado_UpdateZoneOverlayCallback,
			hash          => $hash,
			setting       => $setting,
      zoneID        => $zoneID,
      data          => $myjson
		};

    #Log3 $name, 3, 'NonBlocking Request: ' . Dumper($request);

		HttpUtils_NonblockingGet($request);
        
  }
}



sub TadoAPI_SetAllOverlays(@){
  my ($hash, $setting) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $zonecount = TadoAPI_ZoneRequest($hash);
    my $message = "";
    my $method = "";
    my $myjson ="";

    for (my $i=1; $i <= $zonecount; $i++) {
    	my $URL = $QueryURL . qq{/$homeID/zones/$i/overlay};

      if ($setting eq "off"){
          $method = "DELETE";
          Log3 $name, 3, "TadoAPI $name" . ": " . "Deleting Overlay for Zone $i";
      }elsif($setting > 10){
        $method = "PUT";
        $myjson = { 
          setting => { 
              type => "HEATING",
              power => "ON",
              temperature => {
                  celsius => $setting 
              },
            },
          termination => {
              type => "MANUAL"
            },
          };
          $myjson = encode_json($myjson);
      }

      my $request = {
        url           => $URL,
        header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
        method        => $method,
        timeout       =>  3,
        callback      => \&Tado_UpdateZoneOverlayCallback,
        hash          => $hash,
        setting       => $setting,
        zoneID        => $i,
        data          => $myjson
      };
      HttpUtils_NonblockingGet($request);

    }
    # finaly update readings
    #TadoAPI_UpdateFn($hash);
  }
}

sub TadoAPI_UpdateAllZonesCallback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};

  if($err ne "")
  {
    Log3 $name, 3, "Error in TadoAPI_UpdateZoneCallback while requesting ".$param->{url}." - $err"; 
    return undef;
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
    my $zonecount = TadoAPI_GetZoneCount($hash);
    for (my $i=1; $i <= $zonecount; $i++) {
      Log3 $name, 3, "TadoAPI $name" . ": " . "Set Reading Update for Zone $i ";
      my ($temperature, $humidity, $desiredTemp, $currentHeatingPower, $overlay ) = TadoAPI_GetZoneReadingsById($hash, $i);
      my $zoneName = TadoAPI_GetZoneNameById($hash, $i);

      if(defined($zoneName)){
        readingsBeginUpdate($hash);
        readingsBulkUpdate($hash, "Temperatur_" . $zoneName, $temperature);
        readingsBulkUpdate($hash, "Luftfeuchtigkeit_" . $zoneName, $humidity);
        readingsBulkUpdate($hash, "Heizleistung_" . $zoneName, $currentHeatingPower);
        readingsBulkUpdate($hash, "OverlayType_" . $zoneName, $overlay);
        readingsBulkUpdate($hash, "DesiredTemp_" . $zoneName, $desiredTemp);
        readingsEndUpdate( $hash, 1 );
      }
    }
    readingsBeginUpdate($hash); 
    readingsBulkUpdate($hash, "ActiveZones", $zonecount);
    readingsEndUpdate( $hash, 1 );
  }
}

sub TadoAPI_UpdateFn(@){  
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $TokenData = TadoAPI_LoadToken($hash);
  my $homeID = $attr{$name}{homeID};

  if($apiStatus == 1 && defined($TokenData)){
    # zone specific updates
    my $URL = $QueryURL.qq{/$homeID/zones};
    my $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => 'GET',
			timeout       =>  4,
			hash          => $hash,
      callback      => \&TadoAPI_UpdateAllZonesCallback
		};

		HttpUtils_NonblockingGet($request);
    
    # mobile devices
    $URL=$QueryURL . qq{/$homeID/mobileDevices};
    $request = {
      url           => $URL,
      header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
      method        => 'GET',
      timeout       =>  5,
      callback      => \&TadoAPI_UpdateMobileReadingsCallback,
      hash          => $hash
    };
    HttpUtils_NonblockingGet($request);
    
    # tado devices
    $URL=$QueryURL.qq{/$homeID/zones};
    $request = {
      url           => $URL,
      header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
      method        => 'GET',
      timeout       =>  10,
      callback      => \&TadoAPI_UpdateTadoDeviceInformationCallback,
      hash          => $hash
    };
    HttpUtils_NonblockingGet($request);
   }
}

sub TadoAPI_GetGeoById(@){
  my ($hash, $mobileID) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
	my $URL=$QueryURL.qq{/$homeID/mobileDevices/$mobileID/settings};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $param = {
      url           => $URL,
      header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
      method        => 'GET',
      timeout       =>  2,
      hash          => $hash,
    };

      Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
      my ($err, $data) = HttpUtils_BlockingGet($param);

      if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetGeoById: Error while requesting ".$param->{url}." - $err";
      }
      elsif($data ne "")
      {
        Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
        my $decoded_data = eval { decode_json($data) };
        if ($@){
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetGeoById: decode_json failed, invalid json. error:$@\n";
        }else{
          my $setting = $decoded_data->{'geoTrackingEnabled'};
          readingsBeginUpdate($hash);
          readingsBulkUpdate($hash, "Geolocation_" . $mobileID, $setting);
          readingsEndUpdate( $hash, 1 );
          return $setting;
        }
      }
  }
  return undef;
}
########################################################
# Callback Subs
########################################################
sub TadoAPI_callback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};  
  $param->{code} = 0 unless defined $param->{code};

  if($param->{code} == 401 || $param->{code} == 400){    
      $apiStatus = 1;
      $hash->{STATE}="reachable";
      Log3 $name, 3, "TadoAPI $name" . ": " . "API is reachable. Callback Status: " . $param->{code} if $debug;

  }else{
      $apiStatus = 0;
      $hash->{STATE}="error";
      Log3 $name, 3, "TadoAPI $name" . ": " . "API error: apiStatus $apiStatus ($err)" if $debug;

  }
  return undef;
}

sub Tado_StoreTokenCallback($)
{
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};
  my $tokenFileName = $hash->{TOKEN_FILE};

  if($err ne "")                                                                                                      
  {
    Log3 $name, 3, "TadoAPI $name" . ": " . "StoreTokenCallback: Error in token retrival while requesting ".$param->{url}." - $err";
    $hash->{STATE}="error";
  }

  elsif($data ne "")          
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
    
    my $decoded_data = eval{decode_json($data);};

    if ($@){
      Log3 $name, 3, "TadoAPI $name" . ": " . "StoreTokenCallback: decode_json failed, invalid json. error:$@\n" if $@;
      $hash->{STATE}="error";
    }else{
      #write token data in file
      open(TOKENFILE,">$tokenFileName") or die("ERROR: $!");
      print TOKENFILE $data . "\n";
      close(TOKENFILE);

      # token lifetime management
      $hash->{TOKEN_LIFETIME}  = gettimeofday() + $decoded_data->{'expires_in'};
      $hash->{TOKEN_LIFETIME_HR}  = localtime($hash->{TOKEN_LIFETIME});
      Log3 $name, 3, "TadoAPI $name" . ": " . "StoreTokenCallback: Refreshed authentication token successfully. Valid until " . localtime($hash->{TOKEN_LIFETIME})if $debug;
      $hash->{STATE}="reachable";
    }
  }
}

sub Tado_UpdateZoneOverlayCallback($)
{
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};
  my $zoneID = $param->{zoneID};
  my $setting = $param->{setting};

  if($err ne "")
  {
    Log3 $name, 3, "Error in UpdateZoneOverlayCallback while requesting ".$param->{url}." - $err"; 
    return undef;
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
    Log3 $name, 3, "TadoAPI $name" . ": " . "Set Overlay for Zone $zoneID to: $setting";
    
  }
  # finaly update readings
  TadoAPI_GetZoneReadingsById($hash, $zoneID);

}

sub TadoAPI_LogInfoCallback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};

  if($err ne "")
  {
    Log3 $name, 3, "Error in LogInfoCallback while requesting ".$param->{url}." - $err"; 
    return undef;
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "TadoAPI $name" . ": " . $param->{infotext} .":\n" . $data . "\n";
  }
}

sub TadoAPI_UpdateTadoDeviceInformationCallback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};

  if($err ne "")
  {
    Log3 $name, 3, "Error in UpdateTadoDeviceInformationCallback while requesting ".$param->{url}." - $err"; 
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
    my $decoded_data = eval { decode_json($data) };

    # if api returns error
    eval { my $error = @$decoded_data; };

    if ($@){
      Log3 $name, 3, "TadoAPI $name" . ": " . "UpdateTadoDeviceInformationCallback: decode_json failed, invalid json. error:$@\n" if $@;
      Log3 $name, 3, "TadoAPI $name" . ": " . "UpdateTadoDeviceInformationCallback: Error in decoded data, Code: " . $decoded_data->{'errors'}->[0]->{'code'};
      $hash->{LastRequest}="error";
    }else{
      foreach my $zone ( @$decoded_data ){
        my $devices = $zone->{'devices'};
        foreach my $device ( @$devices ){
          readingsBeginUpdate($hash);
          readingsBulkUpdate($hash, "Battery_" . $device->{'serialNo'}, $device->{'batteryState'});
          readingsEndUpdate( $hash, 1 );
          $hash->{LastRequest}="OK";
        }
      }
    }
  }
}

sub TadoAPI_UpdateMobileReadingsCallback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};

  if($err ne "")
  {
    Log3 $name, 3, "Error in TadoAPI_UpdateMobileReadingsCallback while requesting ".$param->{url}." - $err"; 
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
    my $decoded_data = eval { decode_json($data) };

    # if api returns error
    eval { my $error = @$decoded_data; };

    if ($@){
      Log3 $name, 3, "TadoAPI $name" . ": " . "Decode_json failed, invalid json. error:$@\n" if $@;
      Log3 $name, 3, "TadoAPI $name" . ": " . "Error in TadoAPI_UpdateMobileReadingsCallback, Code: " . $decoded_data->{'errors'}->[0]->{'code'};
      $hash->{LastRequest}="error";
    }else{
      foreach my $item ( @$decoded_data ){
        readingsBeginUpdate($hash);
        readingsBulkUpdate($hash, "Geolocation_" . $item->{'id'}, $item->{'settings'}->{'geoTrackingEnabled'});
        readingsEndUpdate( $hash, 1 );
        $hash->{LastRequest}="OK";
      }
    }
  }
}

sub TadoAPI_GetZoneInfo(@) {
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};   
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    # HomeInfo
    my $URL = qq{https://my.tado.com/api/v2/me};
    my $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => 'GET',
			timeout       =>  3,
      infotext      => "HomeInfos",
			hash          => $hash,
      callback      => \&TadoAPI_LogInfoCallback
		};

		HttpUtils_NonblockingGet($request);

    # TadoDevicesInfo
    $URL = $QueryURL.qq{/$homeID/zones};
    $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => 'GET',
			timeout       =>  3,
      infotext      => "Tado Devices Info",
			hash          => $hash,
      callback      => \&TadoAPI_LogInfoCallback
		};

		HttpUtils_NonblockingGet($request);


    # Mobileinfo
    $URL = $QueryURL . qq{/$homeID/mobileDevices};
    $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => 'GET',
			timeout       =>  3,
      infotext      => "Mobile Devices Info",
			hash          => $hash,
      callback      => \&TadoAPI_LogInfoCallback
		};

		HttpUtils_NonblockingGet($request);


    # todo remove static mobile id
    my $mobileID = $attr{$name}{mobileID};
    $URL=$QueryURL.qq{/$homeID/mobileDevices/$mobileID/settings};
    $request = {
			url           => $URL,
			header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
			method        => 'GET',
			timeout       =>  3,
      infotext      => "Mobile Device $mobileID",
			hash          => $hash,
      callback      => \&TadoAPI_LogInfoCallback
		};

		HttpUtils_NonblockingGet($request);

    # zones
    for (my $i=1; $i <= TadoAPI_GetZoneCount($hash); $i++) {
        $URL=$QueryURL.qq{/$homeID/zones/$i/state};
        my $infotext = "ZoneID $i (" . TadoAPI_GetZoneNameById($hash, $i) . ") Status:";
        my $request = {
          url           => $URL,
          header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
          method        => 'GET',
          timeout       =>  3,
          infotext      => $infotext,
          hash          => $hash,
          callback      => \&TadoAPI_LogInfoCallback
      };

		HttpUtils_NonblockingGet($request);
    }
  }
  return undef;
}

sub TadoAPI_SetGeoByIdCallback($){
  my ($param, $err, $data) = @_;
  my $hash = $param->{hash};
  my $name = $hash->{NAME};

  if($err ne "")
  {
    Log3 $name, 3, "Error in TadoAPI_SetGeoByIdCallback while requesting ".$param->{url}." - $err"; 
  }

  elsif($data ne "")
  {
    Log3 $name, 3, "url ".$param->{url}." returned: $data";
  }

}

sub TadoAPI_SetGeoById(@){
  my ($hash, $mobID, $geo) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
	my $URL=$QueryURL.qq{/$homeID/mobileDevices/$mobID/settings};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    if($geo){
      my $data = { geoTrackingEnabled=>"true" };
    }else{
      my $data = { geoTrackingEnabled=>"false" };
    }

    $data = encode_json($data);

    my $request = {
          url           => $URL,
          header        =>  { "Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}" },
          method        => 'PUT',
          timeout       =>  3,
          mobileID      => $mobID,
          data          => $data,
          hash          => $hash,
          callback      => \&TadoAPI_SetGeoByIdCallback
      };

		HttpUtils_NonblockingGet($request);
  }
}

#######################################################################################################
# API Requests
###############
sub TadoAPI_RequestTadoDevices(@) {
  # returns array with zonenames and zone devices
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my $URL=$QueryURL.qq{/$homeID/zones};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
      my $param = {
      url           => $URL,
      header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
      method        => 'GET',
      timeout       => 2,
      hash          => $hash
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);
    
    if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_RequestTadoDevices: Error while requesting ".$param->{url}." - $err";
      }
      elsif($data ne "")
      {
        Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
        my $decoded_data = eval { decode_json($data) };
        if ($@){
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_RequestTadoDevices: decode_json failed, invalid json. error:$@\n";
        }else{
        my @devices = ();
        foreach my $dev (@$decoded_data){
          push @devices, $dev;
        }
          return @devices;
        }
      }
  }
return undef;
}

sub TadoAPI_ZoneRequest(@) {
  # returns array with state of all zones
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my @zones = ();
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData))
  {
    for (my $i=1; $i <= TadoAPI_GetZoneCount($hash); $i++) {
      my $URL=$QueryURL.qq{/$homeID/zones/$i/state};
      my $param = {
        url           => $URL,
        header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
        method        => 'GET',
        timeout       =>  2,
        hash          => $hash,
        zoneID        => $i
      };

      Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
      my ($err, $data) = HttpUtils_BlockingGet($param);

      if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "ZoneRequest: Error while requesting ".$param->{url}." - $err";
      }
      elsif($data ne "")
      {
          Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;

          # An dieser Stelle die Antwort parsen / verarbeiten mit $data
          my $decoded_data = eval { decode_json($data) };
          if ($@){
            Log3 $name, 3, "TadoAPI $name" . ": " . "RequestZones: Zone $i decode_json failed, invalid json. error:$@\n";
          }else{
            # pushes zone i to array 
            push @zones, $decoded_data;
          }
      }
    }
    return @zones;
  }
}

sub TadoAPI_GetMobileDevices(@) {
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my $URL=$QueryURL . qq{/$homeID/mobileDevices};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $param = {
      url           => $URL,
      header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
      method        => 'GET',
      timeout       => 2,
      hash          => $hash
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);

    if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetMobileDevices: Error while requesting ".$param->{url}." - $err";
      }
    elsif($data ne "")
    {
      my @devices = ();
      Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
      my $decoded_data = eval { decode_json($data) };
      if ($@){
        Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetMobileDevices: decode_json failed, invalid json. error:$@\n";
      }else{
      
        foreach my $item( @$decoded_data ) {
          push @devices, $item;
        }
      }
      return @devices;
    }
  }
return undef;
}

######################################
############ Helpers #################
######################################
sub TadoAPI_ReplaceUmlaute(@) {
  my ($string) = @_;
  my %umlaute = ("ä" => "ae", "Ä" => "Ae", "ü" => "ue", "Ü" => "Ue", "ö" => "oe", "Ö" => "Oe", "ß" => "ss" );
  my $umlautkeys = join ("|", keys(%umlaute));

  $string =~ s/($umlautkeys)/$umlaute{$1}/g;
  return $string;
}

sub TadoAPI_GetHomeId(@){
  # returns first home id only
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData))
  {
    my $param = {
        url           => $DataURL,
        header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
        method        => 'GET',
        timeout       =>  2,
        hash          => $hash,
      };

      Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
      my ($err, $data) = HttpUtils_BlockingGet($param);

      if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "GetHomeId: Error while requesting ".$param->{url}." - $err";
      }
      elsif($data ne "")
      {
        Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
        my $decoded_data = eval { decode_json($data) };
        if ($@){
          Log3 $name, 3, "TadoAPI $name" . ": " . "GetHomeId: Decode_json failed, invalid json. error:$@" if $@;
          $hash->{LastRequest}="error";
        }else{
          $hash->{LastRequest}="OK";
          return $decoded_data->{'homes'}->[0]->{'id'} if (exists($decoded_data->{'homes'}));
        }
      }
  }
  return undef;
}

sub TadoAPI_GetZoneCount(@) {
  my ($hash) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my $URL=$QueryURL.qq{/$homeID/zones};
  my $zonecount = 0;
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $param = {
      url           => $URL,
      header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
      method        => 'GET',
      timeout       => 2,
      hash          => $hash
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);

    if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetZoneCount: Error while requesting ".$param->{url}." - $err";
      }
    elsif($data ne "")
    {
      my @devices = ();
      Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
      my $decoded_data = eval { decode_json($data) };
      #print Dumper($decoded_data);

      if ($@){
        Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetZoneCount: decode_json failed, invalid json. error:$@\n";
      }else{
        return 0 unless (ref($decoded_data) ne 'HASH') ;
        foreach my $item( @$decoded_data ) { 
          $zonecount++;
        }
      
      return $zonecount;
      }
    }
  }
return 0;
}

sub TadoAPI_GetZoneNameById(@) {
  my ($hash, $zoneID) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};
  my $URL=$QueryURL.qq{/$homeID/zones};
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $param = {
      url           => $URL,
      header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
      method        => 'GET',
      timeout       => 2,
      hash          => $hash
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);

    if($err ne "")
      {
          Log3 $name, 3, "TadoAPI $name" . ": " . "GetZoneCount: Error while requesting ".$param->{url}." - $err";
      }
    elsif($data ne "")
    {
      my @devices = ();
      Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
      my $decoded_data = eval { decode_json($data) };

      if ($@){
        Log3 $name, 3, "TadoAPI $name" . ": " . "GetZoneCount: decode_json failed, invalid json. error:$@\n";
      }else{
        if(ref($decoded_data) eq 'ARRAY'){
          my $zoneName = TadoAPI_ReplaceUmlaute(@$decoded_data[$zoneID - 1]->{'name'});
          $zoneName = encode("UTF-8", $zoneName);
          return $zoneName;
        }elsif(ref($decoded_data) eq 'HASH'){
          Log3 $name, 3, "TadoAPI $name" . ": " . "GetZoneCount: " .  $decoded_data->{'errors'}->[0]->{'code'} if (exists($decoded_data->{'errors'}));
        }
      }
    }
  }
  return undef;
}

sub TadoAPI_GetZoneReadingsById(@){  
  my ($hash, $zoneID) = @_;
  my $name = $hash->{NAME};
  my $homeID = $attr{$name}{homeID};   
  my $URL=$QueryURL.qq{/$homeID/zones/$zoneID/state};
	my $temperature = 0;
	my $humidity = 0;
  my $desiredTemp = 0;
  my $currentHeatingPower = 0;
  my $overlay = 0;
  my $TokenData = TadoAPI_LoadToken($hash);

  if(defined($TokenData)){
    my $param = {
        url           => $URL,
        header        => {"Content-Type"=>"application/json;charset=UTF-8","Authorization" => "$TokenData->{'token_type'} $TokenData->{'access_token'}"},
        method        => 'GET',
        timeout       =>  3,
        hash          => $hash
    };

    Log3 $name, 3, 'Blocking GET: ' . Dumper($param) if $debug;
    my ($err, $data) = HttpUtils_BlockingGet($param);

    if($err ne "")
    {
        Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetZoneReadingsById: Error while requesting ".$param->{url}." - $err";
    }
    elsif($data ne "")
    {
      Log3 $name, 3, "url ".$param->{url}." returned: $data" if $debug;
      my $decoded_data = eval { decode_json($data) };
      if ($@){
        Log3 $name, 3, "TadoAPI $name" . ": " . "TadoAPI_GetZoneReadingsById: Zone $zoneID decode_json failed, invalid json. error:$@\n";
      }else{
        $temperature = $decoded_data->{'sensorDataPoints'}->{'insideTemperature'}->{'celsius'};
        $humidity = $decoded_data->{'sensorDataPoints'}->{'humidity'}->{'percentage'};
        $desiredTemp = $decoded_data->{'setting'}->{'temperature'}->{'celsius'};
        $currentHeatingPower = $decoded_data->{'activityDataPoints'}->{'heatingPower'}->{'percentage'};
        $overlay = $decoded_data->{'overlayType'};
        if (!defined $overlay) {$overlay = "no overlay"};

        my $zoneName = TadoAPI_GetZoneNameById($hash, $zoneID);

        readingsBeginUpdate($hash);
        readingsBulkUpdate($hash, "Temperatur_" . $zoneName, $temperature);
        readingsBulkUpdate($hash, "Luftfeuchtigkeit_" . $zoneName, $humidity);
        readingsBulkUpdate($hash, "Heizleistung_" . $zoneName, $currentHeatingPower);
        readingsBulkUpdate($hash, "OverlayType_" . $zoneName, $overlay);
        readingsBulkUpdate($hash, "DesiredTemp_" . $zoneName, $desiredTemp);
        readingsEndUpdate( $hash, 1 );

        return ($temperature, $humidity, $desiredTemp, $currentHeatingPower, $overlay );
      }
    }
  }
  return undef;
}

######################################################
# storePW & readPW Code geklaut aus 96_SIP.pm :)
######################################################
sub TadoAPI_storePassword($$)
{
    my ($name, $password) = @_;
    my $index = "TadoAPI_".$name."_passwd";
    my $key   = getUniqueId().$index;
    my $e_pwd = "";
    
    if (eval "use Digest::MD5;1")
    {
        $key  = Digest::MD5::md5_hex(unpack "H*", $key);
        $key .= Digest::MD5::md5_hex($key);
    }
    
    for my $char (split //, $password)
    {
        my $encode=chop($key);
        $e_pwd.=sprintf("%.2x",ord($char)^ord($encode));
        $key=$encode.$key;
    }

    my $error = setKeyValue($index, $e_pwd);
    return "error while saving TadoAPI password : $error" if(defined($error));
    return "TadoAPI password successfully saved in FhemUtils/uniqueID Key $index";
}

sub TadoAPI_readPassword($)
{
   my ($name) = @_;
   my $index  = "TadoAPI_".$name."_passwd";
   my $key    = getUniqueId().$index;

   my ($password, $error);

   #Log3 $name,5,"$name, read user password from FhemUtils/uniqueID Key $key";
   ($error, $password) = getKeyValue($index);

   if ( defined($error) ) 
   {
      Log3 $name,3, "$name, cant't read Tado password from FhemUtils/uniqueID: $error";
      return undef;
   }  
    
   if ( defined($password) ) 
   {
      if (eval "use Digest::MD5;1") 
      {
         $key  = Digest::MD5::md5_hex(unpack "H*", $key);
         $key .= Digest::MD5::md5_hex($key);
      }

      my $dec_pwd = '';
     
      for my $char (map { pack('C', hex($_)) } ($password =~ /(..)/g)) 
      {
         my $decode=chop($key);
         $dec_pwd.=chr(ord($char)^ord($decode));
         $key=$decode.$key;
      }
      return $dec_pwd;
   }
   else 
   {
      Log3 $name,3,"$name, no Tado password found in FhemUtils/uniqueID";
      return undef;
   }
}

1;

# Beginn der Commandref

=pod
=item [device]
=item summary integration of the Tado API
=item summary_DE Anbindung der Tado Heizungssteuerung über API

=begin html
 -
=end html

=begin html_DE
 -
=end html

# Ende der Commandref
=cut